# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyrimidine',
 'pyrimidine.benchmarks',
 'pyrimidine.learn',
 'pyrimidine.local_search']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.8.0,<4.0.0',
 'numpy>=1.26.0,<2.0.0',
 'pandas>=2.1.0,<3.0.0',
 'scikit-learn>=1.1.0,<2.0.0',
 'scipy>=1.12.0,<2.0.0',
 'toolz>=0.12.0,<0.13.0']

setup_kwargs = {
    'name': 'pyrimidine',
    'version': '1.6',
    'description': 'A competitive framework tailored for GA, crafted with an emphasis on OOP and Algebra-Inspired Programming.',
    'long_description': '# pyrimidine\n\n`pyrimidine` is an extensible framework of genetic/evolutionary algorithm by Python. See [pyrimidine\'s document](https://pyrimidine.readthedocs.io/) for more details.\n\n![LOGO](logo.png)\n\n## Why\n\nWhy is the package named as “pyrimidine”? Because it begins with “py”.\n\n> — Are you kiding?\n>\n> — No, I am serious.\n\n## Download\n\nIt has been uploaded to [pypi](https://pypi.org/project/pyrimidine/), so download it with `pip install pyrimidine`, and also could download it from github.\n\n## Idea\n\nWe regard the population as a container of individuals, an individual as a container of chromosomes\nand a chromosome as a container(array) of genes.\n\nThe container could be a list or an array.\nThe container class has an attribute `element_class`, telling itself the type of the elements in it.\n\nFollowing is the part of the source code of `BaseIndividual` and `BasePopulation`.\n\n```python\nclass BaseIndividual(FitnessModel, metaclass=MetaContainer):\n    element_class = BaseChromosome\n    default_size = 1\n\nclass BasePopulation(PopulationModel, metaclass=MetaContainer):\n    element_class = BaseIndividual\n    default_size = 20\n```\n\nThere is mainly tow kinds of containers: list and tuple as in programming language `Haskell`. See following examples.\n\n```python\n# individual with chromosomes of type _Chromosome\n_Individual1 = BaseIndividual[_Choromosome]\n# individual with 2 chromosomes of type _Chromosome1 and _Chromosome2 respectively\n_Individual2 = MixedIndividual[_Chromosome1, _Chromosome2]\n```\n\n### Math expression\n$s$ of type $S$ is a container of $a:A$, represented as follows:\n\n```\ns={a:A}:S\n```\n\nWe define a population as a container of individuals or chromosomes, and an individual is a container of chromosomes.\n\nAlgebraically, an indivdiual has only one chromosome is equivalent to a chromosome mathematically. A population could also be a container of chromosomes. If the individual has only one chromosome, then just build the population based on chromosomes directly.\n\nThe methods are the functions or operators defined on $s$.\n\n## Use\n\n### Main classes\n\n- BaseGene: the gene of chromosome\n- BaseChromosome: sequence of genes, represents part of a solution\n- BaseIndividual: sequence of chromosomes, represents a solution of a problem\n- BasePopulation: a container of individuals, represents a container of a problem\n                also the state of a stachostic process\n- BaseMultipopulation: a container of population for more complicated optimalization\n\n### import\n\nTo import all algorithms for beginners, simply use the command `from pyrimidine import *`.\n\nTo speed the lib, use the following commands \n```\nfrom pyrimidine import BaseChromosome, BaseIndividual, BasePopulation # import the base classes form `base.py` to build your own classes\n\n# Commands used frequently\nfrom pyrimidine.base import BinaryChromosome, FloatChromosome # import the Chromosome classes and utilize them directly\n# equivalent to `from pyrimidine import BinaryChromosome, FloatChromosome`\nfrom pyrimidine.population import StandardPopulation, HOFPopulation # For creating population with standard GA \n# the same effect with `from pyrimidine import StandardPopulation, HOFPopulation`\nfrom pyrimidine.indiviual import makeIndividual # a helper to make Individual objects, or `from pyrimidine import makeIndividual`\n\nfrom pyrimidine import MultiPopulation # build the multi-populations\nfrom pyrimidine import MetaContainer # meta class for socalled container class, that is recommended to be used for creating novel evolutionary algorithms.\n\nfrom pyrimidine.deco import * # import all decorators\nfrom pyrimidine.deco import fitness_cache, basic_memory # use the cache decorator and memory decorator\n\nfrom pyrimidine import optimize # do optimization implictly with GAs\n```\n\n### subclasses\n\n#### Chromosome\n\nGenerally, it is an array of genes.\n\nAs an array of 0-1s, `BinaryChromosome` is used most frequently.\n\n#### Individual\njust subclass `MonoIndividual` in most cases.\n\n```python\nclass MyIndividual(MonoIndividual):\n    """individual with only one chromosome\n    we set the gene is 0 or 1 in the chromosome\n    """\n    element_class = BinaryChromosome\n\n    def _fitness(self):\n        ...\n```\n\nSince the helper `makeIndividual(n_chromosomes=1, size=8)` could create such individual, it is equivalent to\n\n```python\nclass MyIndividual(binaryIndividual()):\n    # only need define the fitness\n    def _fitness(self):\n        ...\n```\n\n\nIf an individual contains several chromosomes, then subclass `MultiIndividual` or `PolyIndividual`. It could be applied in multi-real-variable optimization problems where each variable has a separative binary encoding.\n\n\nIn most cases, we have to decode chromosomes to real numbers.\n\n```python\nclass _Chromosome(BinaryChromosome):\n    def decode(self):\n        """Decode a binary chromosome\n        \n        if the sequence of 0-1 represents a real number, then overide the method\n        to transform it to a nubmer\n        """\n\nclass ExampleIndividual(BaseIndividual):\n    element_class = _Chromosome\n\n    def _fitness(self):\n        # define the method to calculate the fitness\n        x = self.decode()  # will call decode method of _Chromosome\n        return evaluate(x)\n```\n\n\nIf the chromosomes in an individual are different with each other, then subclass `MixedIndividual`, meanwhile, the property `element_class` should be assigned with a tuple of classes for each chromosome.\n\n```python\nclass MyIndividual(MixedIndividual):\n    """\n    Inherit the fitness from ExampleIndividual directly.\n    It has 6 chromosomes, 5 are instances of _Chromosome, 1 is instance of FloatChromosome\n    """\n    element_class = (_Chromosome,)*5 + (FloatChromosome,)\n```\n\nIt equivalent to `MyIndividual=MixedIndividual[(_Chromosome,)*5 + (FloatChromosome,)]`\n\n#### Population\n\n```python\nclass MyPopulation(StandardPopulation):\n    element_class = MyIndividual\n```\n\nIt is equivalent to `MyPopulation = StandardPopulation[MyIndividual]`.\n\n\n### Initialize randomly\n\n`random` is a factory method!\n\nGenerate a population, with 50 individuals and each individual has 100 genes:\n\n`pop = MyPopulation.random(n_individuals=50, size=100)`\n\nWhen each individual contains 5 chromosomes, use\n\n`pop = MyPopulation.random(n_individuals=10, n_chromosomes=5, size=10)`\n\nHowever, we recommand to set `default_size` in the classes, then run `MyPopulation.random()`\n\n```python\nclass MyPopulation(StandardPopulation):\n    element_class = MyIndividual // 5\n    default_size = 10\n\n# equiv. to\n\nMyPopulation = StandardPopulation[MyIndividual//5]//10\n```\n\nIn fact, `random` method of `BasePopulation` will call random method of `BaseIndividual`. If you want to generate an individual, then just execute `MyIndividual.random(n_chromosomes=5, size=10)`, or set `default_size`, then execute `MyIndividual.random()`.\n\n\n### Evolution\n\n#### `evolve` method\nInitialize a population with `random` method, then call `evolve` method.\n\n```python\npop = MyPopulation.random(n_individuals=50, size=100)\npop.evolve()\nprint(pop.solution)\n```\n\nset `verbose=True` to display the data for each generation.\n\n`evolve` method mainly excute two methods: \n- `init`: initial configuration of the algo.\n- `transition`: do each step of the iteration.\n\n#### History\n\nGet the history of the evolution.\n\n```python\nstat={\'Mean Fitness\':\'mean_fitness\', \'Best Fitness\': lambda pop: pop.best_individual.fitness}\ndata = pop.history(stat=stat)  # use history instead of evolve\n```\n`stat` is a dict mapping keys to function, where string \'mean_fitness\' means function `lambda pop:pop.mean_fitness` which gets the mean fitness of the individuals in `pop`. Since we have defined pop.best_individual.fitness as a property, `stat` could be redefined as `{\'Fitness\': \'fitness\', \'Best Fitness\': \'max_fitness\'}`.\n\nIt requires `ezstat`, a easy statistical tool devoloped by the author.\n\n#### performance\n\nUse `pop.perf()` to check the performance, which calls `evolve` several times.\n\n\n## Example\n\n### Example 1\n\nDescription\n\n    select some of ti, ni, i=1,...,L, ti in {1,2,...,T}, ni in {1,2,...,N}\n    the sum of ni approx. 10, while ti dose not repeat\n\nThe opt. problem is\n\n    min abs(sum_i{ni}-10) + maximum of frequences in {ti}\n    where i is selected.\n\n$$\n\\min_I |\\sum_{i\\in I} n_i -10| + t_m\n\\\\\nI\\subset\\{1,\\cdots,L\\}\n$$\nwhere $t_m$ is the mode of $\\{t_i, i\\in I\\}$\n\n```python\nt = np.random.randint(1, 5, 100)\nn = np.random.randint(1, 4, 100)\n\nimport collections\n\nfrom pyrimidine.individual import makeBinaryIndividual\nfrom pyrimidine.population import StandardPopulation\n\ndef max_repeat(x):\n    # Maximum repetition\n    c = collections.Counter(x)\n    return np.max([b for a, b in c.items()])\n\n\nclass MyIndividual(makeBinaryIndividual()):\n\n    def _fitness(self):\n        x = abs(np.dot(n, self.chromosome)-10)\n        y = max_repeat(ti for ti, c in zip(t, self) if c==1)\n        return - x - y\n\n\nclass MyPopulation(StandardPopulation):\n    element_class = MyIndividual\n\npop = MyPopulation.random(n_individuals=50, size=100)\npop.evolve()\nprint(pop.solution)  # or pop.best_individual.decode()\n```\n\nNote that there is only one chromosome in `MonoIndividual`, which could be got by `self.chromosome`.\n\nIn fact, the population could be the container of chromosomes. Therefore, we can rewrite the classes as follows in a more natural way.\n\n```python\nfrom pyrimidine.chromosome import BinaryChromosome\nfrom pyrimidine.population import StandardPopulation\n\nclass MyChromosome(BinaryChromosome):\n\n    def _fitness(self):\n        x = abs(np.dot(n, self)-10)\n        y = max_repeat(ti for ti, c in zip(t, self) if c==1)\n        return - x - y\n\nclass MyPopulation(StandardPopulation):\n    element_class = MyChromosome\n```\n\nIt is equiv. to\n```python\ndef _fitness(obj):\n    x = abs(np.dot(n, obj)-10)\n    y = max_repeat(ti for ti, c in zip(t, obj) if c==1)\n    return - x - y\n\nMyPopulation = StandardPopulation[BinaryChromosome].set_fitness(_fitness)\n```\n\n### Example2: Knapsack Problem\n\nOne of the famous problem is the knapsack problem. It is a good example for GA.\n\nWe set `history=True` in `evolve` method for the example, that will record the main data of the whole evolution. It will return an object of `pandas.DataFrame`. The argument `stat`  is a dict from a key to function/str(corresponding to a method) representing a mapping from a population to a number. these numbers of one generation will be stored in a row of the dataframe.\n\nsee `# examples/example0`\n\n```python\n#!/usr/bin/env python3\n\nfrom pyrimidine import binaryIndividual, StandardPopulation\nfrom pyrimidine.benchmarks.optimization import *\n\n# generate a knapsack problem randomly\nevaluate = Knapsack.random(n=20)\n\n\nclass MyIndividual(binaryIndividual(size=20)):\n    def _fitness(self):\n        return evaluate(self)\n\nclass MyPopulation(StandardPopulation):\n    element_class = MyIndividual\n    default_size = 10\n\n\npop = MyPopulation.random()\n\nstat={\'Mean Fitness\':\'mean_fitness\', \'Best Fitness\':\'max_fitness\'}\ndata = pop.evolve(stat=stat, history=True) # an instance of `pandas.DataFrame`\n\n# Visualization\nimport matplotlib.pyplot as plt\nfig = plt.figure()\nax = fig.add_subplot(111)\ndata[[\'Mean Fitness\', \'Best Fitness\']].plot(ax=ax)\nax.set_xlabel(\'Generations\')\nax.set_ylabel(\'Fitness\')\nplt.show()\n```\n\n![plot-history](/Users/william/Programming/myGithub/pyrimidine/plot-history.png)\n\n\n## Extension\n\n`pyrimidine` is extremely extendable. It is easy to implement other iterative models or algorithms, such as simulation annealing(SA) and particle swarm optimization(PSO).\n\nCurrently, it is recommended to define subclasses based on `IterativeModel` as a mixin. (not mandatory)\n\nIn PSO, we regard a particle as an individual, and `ParticleSwarm` as a population. But in the following, we subclass it from `IterativeModel`\n\n```python\n# pso.py\n@basic_memory\nclass Particle(BaseIndividual):\n    """A particle in PSO\n\n    Extends BaseIndividual\n\n    Variables:\n        default_size {number} -- one individual represented by 2 chromosomes: position and velocity\n        phantom {Particle} -- the current state of the particle moving in the solution space.\n    """\n\n    element_class = FloatChromosome\n    default_size = 2\n\n    # other methods\n\nclass ParticleSwarm(PopulationMixin):\n    """Standard PSO\n    \n    Extends:\n        PopulationMixin\n    """\n\n    element_class = Particle\n    default_size = 20\n\n    params = {\'learning_factor\': 2, \'acceleration_coefficient\': 3,\n    \'inertia\':0.75, \'n_best_particles\':0.2, \'max_velocity\':None}\n\n    def init(self):\n        for particle in self:\n            particle.init()\n        self.hall_of_fame = self.get_best_individuals(self.n_best_particles, copy=True)\n\n    def update_hall_of_fame(self):\n        hof_size = len(self.hall_of_fame)\n        for ind in self:\n            for k in range(hof_size):\n                if self.hall_of_fame[-k-1].fitness < ind.fitness:\n                    self.hall_of_fame.insert(hof_size-k, ind.copy())\n                    self.hall_of_fame.pop(0)\n                    break\n\n    @property\n    def best_fitness(self):\n        if self.hall_of_fame:\n            return max(map(attrgetter(\'fitness\'), self.hall_of_fame))\n        else:\n            return super().best_fitness\n\n    def transition(self, *args, **kwargs):\n        """\n        Transitation of the states of particles\n        """\n        self.move()\n        self.backup()\n        self.update_hall_of_fame()\n\n    def backup(self):\n        # overwrite the memory of the particle if its current state is better its memory\n        for particle in self:\n            particle.backup(check=True)\n\n    def move(self):\n        """Move the particles\n\n        Define the moving rule of particles, according to the hall of fame and the best record\n        """\n\n        scale = random()\n        eta = random()\n        scale_fame = random()\n        for particle in self:\n            for fame in self.hall_of_fame:\n                if particle.fitness < fame.fitness:\n                    particle.update_vilocity_by_fame(fame, scale, scale_fame, \n                        self.inertia, self.learning_factor, self.acceleration_coefficient)\n                    particle.position = particle.position + particle.velocity\n                    break\n        for particle in self.hall_of_fame:\n            particle.update_vilocity(scale, self.inertia, self.learning_factor)\n            particle.position = particle.position + particle.velocity\n```\n\nIf you want to apply PSO, then you can define\n\n```python\nclass MyParticleSwarm(ParticleSwarm, BasePopulation):\n    element_class = _Particle\n    default_size = 20\n\npop = MyParticleSwarm.random()\n```\n\nOf course, it is not mandatory. It is allowed to inherit `ParticleSwarm` from for example `HOFPopulation` directly.\n',
    'author': 'William Song',
    'author_email': '30965609+Freakwill@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Freakwill/pyrimidine',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

"""
Tabu Search was created by Fred W. Glover in 1986 and formalized in 1989

*References*
Glover, Fred W. and Manuel Laguna. “Tabu Search.” (1997).
Glover, Fred W.. “Tabu Search - Part I.” INFORMS J. Comput. 1 (1989): 190-206.
Glover, Fred W.. “Tabu Search - Part II.” INFORMS J. Comput. 2 (1989): 4-32.
"""

from ..base import BaseIndividual
from ..utils import random, choice
from ..deco import basic_memory


@basic_memory
class BaseTabuSearch(BaseIndividual):
    """Tabu Search algorithm
    """

    params = {'value': 0,
        'tabu_list': [],
        'actions': [],
        'tabu_size': 10
        }

    def transition(self, *args, **kwargs):
        action = choice(self.actions)
        cpy = self.get_neighbour(action)
        if action not in self.tabu_list:
            if cpy.fitness > self.fitness:
                self.chromosomes = cpy.chromosomes
                self.set_memory({
                    'fitness': cpy.fitness,
                    'solution': cpy.decode()
                    })
            else:
                if random() < 0.02:
                    self.chromosomes = cpy.chromosomes
                    self.set_memory({
                        'fitness': cpy.fitness,
                        'solution': cpy.decode()
                        })
                else:
                    self.tabu_list.append(action)
        else:
            if cpy.fitness > self.fitness:
                self.chromosomes = cpy.chromosomes
                self.set_memory({
                    'fitness': cpy.fitness,
                    'solution': cpy.decode()
                    })
                self.tabu_list.remove(action)
        self.update_tabu_list()

    def update_tabu_list(self):
        if len(self.tabu_list) > self.tabu_size:
            self.tabu_list.pop(0)

    def get_neighbour(self, action):
        raise NotImplementedError


class SimpleTabuSearch(BaseTabuSearch):

    def get_neighbour(self, action):
        cpy = self.copy()
        i, j = action
        cpy.chromosomes[i][j] = cpy.gene.random()
        return cpy



#!/usr/bin/env python3

"""Variants of Population classes

A population is defined as a container of individuals.

StandardPopulation: Standard Genetic Algorithm
HOFPopulation: Standard Genetic Algorithm with hall of fame
"""

from operator import methodcaller, attrgetter
from random import gauss, random

import numpy as np

from .base import BasePopulation, BaseIndividual
from .individual import binaryIndividual
from .chromosome import BinaryChromosome
from .meta import MetaList

from .deco import side_effect


class StandardPopulation(BasePopulation):
    """Standard GA
    
    Extends:
        BasePopulation

    Params:
        n_elders: the number (or rate) of the last generation
    """

    params = {'n_elders': 0.5}
    
    def transition(self, *args, **kwargs):
        """
        Transitation in Standard GA
        """
        elder = self.get_best_individuals(self.n_elders, copy=True)
        super().transition(*args, **kwargs)
        self.extend(elder)


Population = StandardPopulation


class HOFPopulation(BasePopulation):
    """GA with hall of fame
    
    Extends:
        StandardPopulation
    
    Attributes:
        hall_of_fame (list): the copy of the best individuals
    """

    params = {'hof_size': 2}
    alias ={'hof': 'hall_of_fame'}
    
    # hall_of_fame = []

    def init(self):
        super().init()
        self.hall_of_fame = self.get_best_individuals(self.hof_size, copy=True)

    def transition(self, *args, **kwargs):
        """
        Update the `hall_of_fame` after each step of evolution
        """

        self.mate_with(self.hall_of_fame)
        super().transition(*args, **kwargs)
        self.update_hall_of_fame()

    def update_hall_of_fame(self):
        """Update the hall of fame

        insert the best individuals of the population into the hall of fame;
        meanwhile, remove the worst ones in the hall of fame
        """
        for ind in self:
            for k in range(self.hof_size):
                if self.hall_of_fame[-k-1].fitness < ind.fitness:
                    self.hall_of_fame.insert(self.hof_size-k, ind.copy())
                    self.hall_of_fame.pop(0)
                    break

    @property
    def max_fitness(self):
        if self.hall_of_fame:
            return max(map(attrgetter('fitness'), self.hall_of_fame))
        else:
            return super().max_fitness

    @property
    def best_individual(self):
        if self.hall_of_fame:
            return self.hall_of_fame[np.argmax([_.fitness for _ in self.hall_of_fame])]
        else:
            return super().best_individual


class DualPopulation(StandardPopulation):
    """Dual Genetic Algo.
    
    Extends:
        StandardPopulation
    """

    params ={
    'dual_prob': 0.2,
    'n_elders': 0.3}

    def dual(self):
        for k, ind in enumerate(self):
            if random() < self.dual_prob:
                d = ind.dual()
                if d.fitness > ind.fitness:
                    self[k] = d
    
    def transition(self, *args, **kwargs):
        self.dual()
        super().transition(*args, **kwargs)


class GamogenesisPopulation(HOFPopulation):
    """Gamogenesis Genetic Algo.
    
    Extends:
        HOFPopulation
    """

    def mate(self, mate_prob=None):
        """Mate the male sub-population and female sub-population.

        Keyword Arguments:
            mate_prob {number} -- the proba. of mating of two individuals (default: {None})
        """

        mate_prob = mate_prob or self.mate_prob
        offspring = [individual.cross(other) for individual, other in zip(self.males, self.females)
        if random() < mate_prob]
        self.individuals.extend(offspring)
        return offspring

    def get_homosex(self, x=0):
        return [i for i in self if i.gender==x]


class EliminationPopulation(BasePopulation):

    def transition(self, *args, **kwargs):
        elder = self.clone()
        elder.select()
        super().transition(*args, **kwargs)
        self.eliminate()
        self.merge(elder)

    def eliminate(self):
        # remove some individuals randomly from the population
        for individual in self:
            if random() < individual.eliminate_prob():
                self.remove(individual)


class AgePopulation(EliminationPopulation):

    def transition(self, *args, **kwargs):
        for individual in self:
            individual.age += 1
        super().transition(*args, **kwargs)

    def eliminate(self):
        # remove some old individuals
        for individual in self:
            if random() * individual.life_span < individual.age:
                self.remove(individual)


class LocalSearchPopulation(StandardPopulation):
    """Population with `local_search` method
    """

    params = {'n_local_iter': 2}

    # def init(self):
    #     for i in self:
    #         i.n_iter = self.n_local_iter

    def transition(self, *args, **kwargs):
        """Transitation of the states of population

        Calling `local_search` method
        """
        super().transition(*args, **kwargs)
        self.local_search(n_iter=self.n_local_iter)


class ModifiedPopulation(StandardPopulation):

    params = {'mutate_prob_ub':0.5, 'mutate_prob_lb':0.1}

    def mutate(self):
        fm = self.max_fitness
        fa = self.mean_fitness
        for individual in self:
            f = individual.fitness
            if f > fa:
                mutate_prob = self.mutate_prob_ub - (self.mutate_prob_ub-self.mutate_prob_lb) * (f-fa) / (fm-fa)
            else:
                mutate_prob = self.mutate_prob_ub
            if random() < mutate_prob:
                individual.mutate()


def makeBinaryPopulation(n_individuals=20, size=8, as_chromosome=True, cls=None):
    """Make a binary population
    
    Args:
        n_individuals (int, optional): the number of the individuals in the population
        size (int, optional): the size of an individual or a chromosome
        as_chromosome (bool, optional): take chromosomes as the individuals of the population
        cls (None, optional): the type of the population (HOFPopulation by default)
    
    Returns:
        subclass of BasePopulation
    """

    cls = cls or HOFPopulation
    if as_chromosome:
        return cls[BinaryChromosome // size] // n_individuals
    else:
        return cls[binaryIndividual(size)] // n_individuals



#!/usr/bin/env python3

import numpy as np

from ..utils import prufer_decode
from ..benchmarks import BaseProblem


class Knapsack(BaseProblem):
    """Knapsack Problem

    max sum_i ci xi
    s.t. sum_i wi xi <= W
    xi = 0 / 1 (choice variable)
    where ci is in c, wi is the coresponding weight of ci
    """
    def __init__(self, w, c, W=0.7, M=100):
        """
        
        Arguments:
            w {array} -- weight array of goods
            c {array} -- value array of goods
            W {number} -- upper bound (proportion) of total weight
        
        Keyword Arguments:
            M {number} -- penalty (default: {100})
        """

        if W < 1:
            W = np.sum(w) * W

        self.w = w
        self.c = c
        self.W = W
        self.M = M
        self.__sorted = None

    @property
    def n_bags(self):
        return len(self.c)

    @staticmethod
    def random(n_bags=50, W=0.7):
        w = np.random.randint(1, 21, n_bags)
        c = np.random.randint(1, 21, n_bags)
        return Knapsack(w, c, W=W)

    @staticmethod
    def example(W=0.7):
        w = [71, 34, 82, 23, 1, 88, 12, 57, 10, 68, 5, 33, 37, 69, 98, 24, 26, 83, 16, 26, 8]
        c = [26, 59, 30, 19, 66, 85, 94, 8, 3, 44, 5, 1, 41, 82, 76, 1, 12, 81, 73, 32, 3]
        return Knapsack(w, c, W=W)

    def argsort(self):
        return np.argsort(self.c / self.w)

    @property
    def sorted(self):
        if self.__sorted is None:
            self.__sorted =self.argsort()

        return self.__sorted

    def __call__(self, x):
        # x is a binary array
        c, w, W, M = self.c, self.w, self.W, self.M
        v = np.sum([ci for i, ci in zip(x, c) if i==1])
        w = np.sum([wi for i, wi in zip(x, w) if i==1])
        if  w <= W:
            return v
        else:
            return - 1/(1 + np.exp(-v)) * M

        # a = np.array([1,2,3,4,5,6,1,1,1,1])
        # return np.sum(a[x==1])

class MultiKnapsack(BaseProblem):
    """Multi Choice Knapsack Problem

    max sum_ij cij xij
    s.t. sum_ij wij xij <= W
    {xij} is 
    where xij is the choice variable, that is sum_j xij=1
    """

    def __init__(self, ws, cs, W=0.7, M=100):
        """
        Arguments:
            ws {tuple of arrays} -- weight array of goods
            cs {tuple of arrays} -- value array of goods
            W {number} -- upper bound (proportion) of total weight
        
        Keyword Arguments:
            M {number} -- penalty (default: {100})
        """

        if W < 1:
            W = sum(map(np.sum, ws)) * W

        self.ws = ws
        self.cs = cs
        self.W = W
        self.M = M
        self.__sorted = None
        self.size = tuple(map(len, cs))

    @staticmethod
    def random(size=(4,5,2,6,5,7,8,3), W=0.7):
        ws = tuple(np.random.randint(1, 51, n) for n in size)
        cs = tuple(np.random.randint(1, 21, n) for n in size)
        return MultiKnapsack(ws, cs, W=W)

    def argsort(self):
        c = min(self.cs)
        w = min(self.ws)
        return np.argsort(c / w)

    def __call__(self, x):
        # x is an integer array
        assert all(0<=xi < s for xi, s in zip(x, self.size))
        c, w, W, M = self.c, self.w, self.W, self.M
        v = np.sum([ci[xi] for xi, ci in zip(x, c)])
        w = np.sum([wi[xi] for xi, wi in zip(x, w)])
        if  w <= W:
            return v
        else:
            return - 1/(1 + np.exp(-v)) * M


class MLE(BaseProblem):
    # max likelihood estimate

    def __init__(self, pdf, x):
        self.pdf = logpdf
        self.x = x

    @staticmethod
    def random(size=300):
        from scipy.stats import norm
        x = norm.rvs(size=size)
        return MLE(logpdf=norm.logpdf, x=x)

    def __call__(self, t):
        return np.sum([self.logpdf(xi, *t) for xi in self.x])


class MixMLE(BaseProblem):
    # mix version of max likelihood estimate
    # x|k ~ pk
    def __init__(self, pdfs, x):
        self.pdfs = pdfs
        self.x = x

    @staticmethod
    def random(n_observants=300, n_components=2):
        from scipy.stats import norm
        x1 = norm.rvs(size=n_observants)
        x2 = norm.rvs(loc=2, size=n_observants)
        x = np.hstack((x1, x2))
        return MixMLE(pdfs=(norm.pdf,)*n_components, x=x)

    def logpdf(self, x, t, a):
        return np.log(np.dot([pdf(x, ti) for pdf, ti in zip(self.pdfs, t)], a))

    def __call__(self, t, a):
        # assert np.sum(a) == 1
        return np.sum([self.logpdf(xi, t, a) for xi in self.x])


from scipy.spatial.distance import pdist, squareform

class ShortestPath(BaseProblem):
    def __init__(self, points):
        """TSP
        
        Arguments:
            points {array with shape of N * 2} -- points of the path
        """
        self.points = points
        self._dm = squareform(pdist(points))

    @staticmethod
    def random(N):
        return ShortestPath(np.random.random(size=(N, 2)))

    def __call__(self, x):
        return np.sum([self._dm[i,j] if i<j else self._dm[j, i] for i, j in zip(x[:-1], x[1:])])


class CurvePath(ShortestPath):
    def __init__(self, x, y):
        _points = np.column_stack([x,y])
        super().__init__(_points)


def _heart(t, a=1):
    x = np.cos(t)
    y = np.sin(t) + np.power(x**2, 1/3)
    return a * x, a * y

t = np.linspace(0, 2*np.pi, 50)
x1, y1 = _heart(t)
x2, y2 = _heart(t, a=1.3)
x = np.hstack((x1, x2))
y = np.hstack((y1, y2))
heart_path = CurvePath(x, y)


class MinSpanningTree(BaseProblem):
    def __init__(self, nodes, edges=[], weights=None):
        self.nodes = nodes
        self.edges = edges
        self.weights = weights

    def __call__(self, x):
        edges = prufer_decode(x, self.nodes)
        return sum(self.weights[edge] for edge in edges)


class FacilityLayout(BaseProblem):
    """
    F: F
    D: D
    """
    
    def __init__(self, F, D):
        self.F = F
        self.D = D
    
    @staticmethod
    def random(self, n):
        F = np.random.random(size=(n, n))
        D = np.random.random(size=(n, n))
        return FacilityLayout(F, D)

    def __call__(self, x):
        return np.dot(self.F.ravel(), np.array([self.D[xi, xj] for xj in x for xi in x]))



#!/usr/bin/env python3


from .chromosome import QuantumChromosome

from .deco import basic_memory

@basic_memory
class _Individual(QuantumChromosome):

    def mutate(self):
        pass

    def cross(self, other):
        return self.__class__((self + other) /2)


class QuantumPopulation(HOFPopulation):

    element_class = _Individual
    default_size = 20

    def backup(self):
        for i in self: i.backup()

    def update_hall_of_fame(self, *args, **kwargs):
        """
        Update the `hall_of_fame` after each step of evolution
        """

        self.backup()
        super().update_hall_of_fame(*args, **kwargs)



#!/usr/bin/env python3


import numpy as np
import numpy.linalg as LA

from ..benchmarks import BaseProblem

_basis = [lambda x: np.ones(len(x)), lambda x: x, lambda x: x**2,
np.sin, np.cos, np.tan,
np.exp, lambda x: np.log(np.abs(x)+1), lambda x:x>0]

n_basis_ = len(_basis)


def lin_comb(x, coefs, basis):
    return np.sum([c*b(x) for c, b in zip(coefs, basis)], axis=0)


class Function1DApproximation(BaseProblem):

    def __init__(self, function, lb=0, ub=1, basis=_basis):
        self.function = function
        self.lb = lb
        self.ub = ub
        self.x = np.linspace(self.lb, self. ub, 30)
        self.y = self.function(self.x)
        self.threshold = 1
        self.basis = basis

    def __call__(self, coefs):
        """A: N * K
        C: K
        B: K*p
        """

        yy = np.sum([c*b(self.x) for c, b in zip(coefs, self.basis)], axis=0)
        p = 0.001
        return - np.mean(np.abs((yy - self.y) / (np.abs(self.y) + 1))) + p*np.mean(np.abs(coefs) < self.threshold)



#!/usr/bin/env python

"""
(mu + lambda) - Evolution Strategy

*References*
Rechenberg, I. 1973. Evolutionsstrategie – Optimierung technischer Systeme nach Prinzipien der biologischen Evolution, Frommann-Holzboog.
"""


from .base import BasePopulation
from .utils import randint2


class EvolutionStrategy(BasePopulation):
    # Evolution Strategy

    params ={
        "mu" : 10,
        "lambda_": 20,
    }

    def init(self):
        super().init()
        if 'mu' not in self.params:
            self.set_params(mu=self.default_size) 

    def transition(self, *args, **kwargs):
        offspring = self.mate()
        self.extend(offspring)
        self.mate()
        self.select_best_individuals(self.mu)

    def mate(self, lambda_=None):
        lambda_ = lambda_ or self.lambda_
        offspring = []
        n = len(self)
        for _ in range(lambda_):
            i, j = randint2(0, n-1)
            child = self[i].cross(self[j])
            offspring.append(child)
        return offspring

    def select_best_individuals(self, mu=None):
        mu = mu or self.mu
        self.individuals = self.get_best_individuals(mu)
